#include <proto.h>

#include "gtmanager.h"
#include "gterror.h"

/************************************************************************\
*
* GtManager class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

/************************************************************************\
*
* Create and init manager.
*
\************************************************************************/
GtManager *GtMkManager(char *n,GtManager *p,GtTag *tgs)
{
	GtManager *obj;
	
	if(!(obj=new GtManager()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtManager object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init manager.
*
\************************************************************************/
GtManager *GtMkManager(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtManager *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkManager(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}
	
/************************************************************************\
*
* Setup empty manager.
*
\************************************************************************/
GtManager::GtManager(void)
{
	GtSInit(&unman);
	GtSInit(&man);
	maxchild=-1;
	nrchild=0;
	haswless=FALSE;
}

/************************************************************************\
*
* Destroy manager.
*
\************************************************************************/
void GtManager::Destroy(void)
{
	GtWList *wl;
	
	/* Delete all unmanaged widgets first since they may destroy childrean */
	/* of this object that they control. */
	this->Map(FALSE);
	while((wl=(GtWList *)GtSFirst(&unman)))
	{
		GtDestroy(wl->w);
		free(wl);
	}
	while((wl=(GtWList *)GtSFirst(&man)))
	{
		GtDestroy(wl->w);
		free(wl);
	}
	GtWidget::Destroy();
}

/************************************************************************\
*
* Set manager attributes.
*
\************************************************************************/
sbool GtManager::Set(GtTag *tgs)
{
	GtTag *t;
	GtWList *wl;
	sbool dch=FALSE;
	
	if(!GtWidget::Set(tgs))
		return(FALSE);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Disabled:
				dch=TRUE;
				break;
		}
	}
	if(dch)
	{
		for(wl=(GtWList *)GtSFirst(&man); wl;
				wl=(GtWList *)GtSNext(&wl->lnk))
		{
			wl->w->SetV(GT_Disabled,flags & GTWF_DISABLED,GT_DONE);
		}
	}
	return(TRUE);
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtManager::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtManager")==0)
		return(TRUE);
	else
		return(GtWidget::IsSubClassOf(cn));
}

/************************************************************************\
*
* Make sure no child of an unmapped parent is the focus.
*
\************************************************************************/
void GtManager::Map(sbool map)
{
	GtWidget::Map(map);
	if(!(flags & GTWF_MAPPED))
		dsp->ChkFocus(this);
}

/************************************************************************\
*
* Build manager window and child windows.
*
\************************************************************************/
sbool GtManager::Build(void)
{
	GtWList *wl;
	
	for(wl=(GtWList *)GtSFirst(&unman); wl;
			wl=(GtWList *)GtSNext(&wl->lnk))
	{
		if(!wl->w->Build())
			return(FALSE);
	}
	for(wl=(GtWList *)GtSFirst(&man); wl;
			wl=(GtWList *)GtSNext(&wl->lnk))
	{
		if(!wl->w->Build())
			return(FALSE);
	}
	if(!GtWidget::Build())
		return(FALSE);
	this->Relayout();
	flags&=~GTWF_ONCE;
	if(!parent || !(parent->flags & GTWF_UNBUILT))
		this->Layout();
	return(TRUE);
}

/************************************************************************\
*
* Refresh manager and it's windowless childrean.
*
\************************************************************************/
void GtManager::Refresh(void)
{
	GtWList *wl;
	
	GtWidget::Refresh();
	if(gfx)
	{
		if(haswless)
		{
			for(wl=(GtWList *)GtSFirst(&man); wl;
					wl=(GtWList *)GtSNext(&wl->lnk))
			{
				if(wl->nowin)
					wl->w->Refresh();
			}
		}
	}
}

/************************************************************************\
*
* Handle own events and all child events.
*
\************************************************************************/
int GtManager::Handle(XEvent *xe)
{
	GtWList *wl;
	int ret;
	
	if((flags & GTWF_NOWIN) && !(flags & GTWF_MAPPED))
		return(GTHR_SEARCH);
	ret=GtWidget::Handle(xe);
	if((ret!=GTHR_SEARCH) && (xe->type==Expose))
	{
		if(haswless)
		{
			for(wl=(GtWList *)GtSFirst(&man); wl;
					wl=(GtWList *)GtSNext(&wl->lnk))
			{
				if(wl->nowin)
					wl->w->Handle(xe);
			}
		}
	}
	return(ret);
}

/************************************************************************\
*
* Change position of child in parant's list.
*
\************************************************************************/
void GtManager::PositionChild(GtWidget *o,ulong p)
{
	GtWList *wl;
	
	if(!o)
		return;
	if(o->Manage())
	{
		for(wl=(GtWList *)GtSFirst(&man); wl;
				wl=(GtWList *)GtSNext(&wl->lnk))
		{
			if(wl->w==o)
				break;
		}
		if(wl)
		{
			GtSRemove(&man,&wl->lnk);
			GtSAddAt(&man,&wl->lnk,p);
			if(!(flags & GTWF_UNBUILT))
				this->Relayout();
		}
	}
	else
	{
		for(wl=(GtWList *)GtSFirst(&man); wl;
				wl=(GtWList *)GtSNext(&wl->lnk))
		{
			if(wl->w==o)
				break;
		}
		if(wl)
		{
			GtSRemove(&man,&wl->lnk);
			GtSAddAt(&man,&wl->lnk,p);
		}
	}
}
		
/************************************************************************\
*
* Add child object to manager.
*
\************************************************************************/
sbool GtManager::AddChild(GtWidget *o,ulong p)
{
	GtWList *wl;
	
	if(!o)
		return(TRUE);
	if(o->Manage())
	{
		if((maxchild<0) || (nrchild<maxchild))
			nrchild++;
		else
		{
			GtError(this,GTEC_NOADD,"Unable to add child: %s\n",o->GetName());
			return(FALSE);
		}
		if(!(wl=this->MakeWList(o,0)))
			return(FALSE);
		GtSAddAt(&man,&wl->lnk,p);
		flags|=GTWF_CHG;
		if(!(flags & GTWF_UNBUILT))
			this->Relayout();
	}
	else
	{
		if(!(wl=(GtWList *)malloc(sizeof(GtWList))))
		{
			GtError(this,GTEC_NOMEMORY,"No memory for widget list!\n");
			return(FALSE);
		}
		wl->w=o;
		wl->nowin=FALSE;
		GtSAddAt(&unman,&wl->lnk,p);
	}
	return(TRUE);
}

/************************************************************************\
*
* Remove child object from manager.
*
\************************************************************************/
void GtManager::RemChild(GtWidget *o)
{
	GtWList *wl=NULL,*tmp;
	
	if(o->Manage())
	{
		haswless=FALSE;
		for(tmp=(GtWList *)GtSFirst(&man); tmp;
				tmp=(GtWList *)GtSNext(&tmp->lnk))
		{
			if(tmp->w==o)
				wl=tmp;
			else
				haswless=haswless || tmp->nowin;
		}
		if(wl)
		{
			GtSRemove(&man,&wl->lnk);
			if(!(flags & GTWF_UNBUILT))
				this->Relayout();
		}
	}
	else
	{
		for(wl=(GtWList *)GtSFirst(&unman); wl;
				wl=(GtWList *)GtSNext(&wl->lnk))
		{
			if(wl->w==o)
				break;
		}
		if(wl)
			GtSRemove(&unman,&wl->lnk);
	}	
}

/************************************************************************\
*
* Set attributes of named child.
*
\************************************************************************/
sbool GtManager::NamedSet(char *n,GtTag *tgs)
{
	GtWList *wl;
	
	for(wl=(GtWList *)GtSFirst(&unman); wl;
			wl=(GtWList *)GtSNext(&wl->lnk))
	{
		if(strcmp(n,wl->w->GetName())==0)
			return(wl->w->Set(tgs));
	}
	for(wl=(GtWList *)GtSFirst(&man); wl;
			wl=(GtWList *)GtSNext(&wl->lnk))
	{
		if(strcmp(n,wl->w->GetName())==0)
			return(wl->w->Set(tgs));
	}
	return(FALSE);
}

/************************************************************************\
*
* Set attributes of named child.
*
\************************************************************************/
sbool GtManager::NamedSetV(char *n,ulong tag,...)
{
	va_list va;
	int ps;
	sbool res;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	res=this->NamedSet(n,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(res);
}

/************************************************************************\
*
* Move and resize child on request.
*
\************************************************************************/
void GtManager::MoveSizeReq(GtWidget *o,word xp,word yp,word xs,word ys)
{
	o->SetPosSize(xp,yp,xs,ys);
}

/************************************************************************\
*
* Find named child.
*
\************************************************************************/
GtWidget *GtManager::Find(char *n)
{
	GtWList *wl;
	GtWidget *ret;
	
	if(strcmp(n,name)==0)
		return(this);
	for(wl=(GtWList *)GtSFirst(&unman); wl;
			wl=(GtWList *)GtSNext(&wl->lnk))
	{
		if((ret=wl->w->Find(n)))
			return(ret);
	}
	for(wl=(GtWList *)GtSFirst(&man); wl;
			wl=(GtWList *)GtSNext(&wl->lnk))
	{
		if((ret=wl->w->Find(n)))
			return(ret);
	}
	return(NULL);
}

/************************************************************************\
*
* Add exposure mask to parent if it has windowless childrean.
*
\************************************************************************/
void GtManager::SetEMask(ulong em)
{
	GtWidget::SetEMask(haswless ? em|ExposureMask : em);
}

/************************************************************************\
*
* Set minimum and maximum size and layout if any change.
*
\************************************************************************/
sbool GtManager::SetMinMaxSize(word xs,word ys,word mxs,word mys)
{
	sbool ret;
	
	ret=GtWidget::SetMinMaxSize(xs,ys,mxs,mys);
	flags&=~GTWF_CHG;
	if(!(flags & GTWF_ONCE))
		this->Layout();
	return(ret);
}

/************************************************************************\
*
* Recursive remap of colors and colormap.
*
\************************************************************************/
void GtManager::RemapColors(uword msk,ulong bg,ulong fg,
										ulong lt,ulong dk,ulong fl,ulong ds,GtColormap *cmap)
{
	GtWList *wl;
	
	GtWidget::RemapColors(msk,bg,fg,lt,dk,fl,ds,cmap);
	for(wl=(GtWList *)GtSFirst(&man); wl;
			wl=(GtWList *)GtSNext(&wl->lnk))
	{
		wl->w->RemapColors(msk,bg,fg,lt,dk,fl,ds,cmap);
	}
}

/************************************************************************\
*
* Set new position and size and layout.
*
\************************************************************************/
void GtManager::SetPosSize(word xp,word yp,word xs,word ys)
{
	word oxs,oys;
	
	oxs=xsize;
	oys=ysize;
	GtWidget::SetPosSize(xp,yp,xs,ys);
	if((oxs!=xsize) || (oys!=ysize) || (flags & GTWF_NOWIN))
		this->Layout();
}

/************************************************************************\
*
* Create widget list with layout information.
*
\************************************************************************/
GtWList *GtManager::MakeWList(GtWidget *o,int sz)
{
	GtWList *wl;
	
	if(!(wl=(GtWList *)malloc(sizeof(GtWList)+sz)))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for layout information!\n");
		return(FALSE);
	}
	memset(wl,0,sizeof(GtWList)+sz);
	GtSInit(&wl->lnk);
	wl->w=o;
	wl->nowin=o->IsWindowLess();
	o->SetLInfo(wl);
	haswless=haswless || wl->nowin;
	return(wl);
}
